# RAGU - Radar Analysis Graphical Utility
#
# copyright © 2020 btobers <tobers.brandon@gmail.com>
#
# distributed under terms of the GNU GPL3.0 license
""" 
constants
"""
### imports ###

# some physical constants for Maxwell's equation for speed of light in a dielectric medium
C = 299792458                   # speed of light in a vacuum
# Eps_0 = 8.8541878 * 10**(-12)   # epsilon naught (vacuum permittivity)
# Mu_0 = 1.257 * 10**(-6)         # mu naught (vacuum permeability)